# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains various utility functions that
are used liberally throughout matador.

"""


__all__ = []
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"
