# coding: utf-8
# Distributed under the terms of the MIT License.

""" The calculator module contains classes for use as DFT/atomistic
calculators within the compute module class.

"""


__all__ = ["CastepCalculator"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .castep import CastepCalculator
